<?php

namespace jrosset\CliProgram;

use FilesystemIterator;
use ReflectionClass;
use ReflectionException;
use Symfony\Component\Console\Command\Command;

/**
 * A spot for command auto discovery
 */
class AutoDiscoverySpot implements IAutoDiscoverySpot {
    /**
     * @var string The directory path
     */
    private string $directoryPath;
    /**
     * @var bool True if search in subdirectories too, else False
     */
    private bool $processSubDirectories;

    /**
     * Initialization
     *
     * @param string $directoryPath         The directory path
     * @param bool   $processSubDirectories True if search in subdirectories too, else False
     */
    public function __construct (string $directoryPath, bool $processSubDirectories = true) {
        $this->setDirectoryPath($directoryPath);
        $this->setProcessSubDirectories($processSubDirectories);
    }

    /**
     * The directory path
     *
     * @return string The directory path
     */
    public function getDirectoryPath (): string {
        return $this->directoryPath;
    }
    /**
     * Set the directory path
     *
     * @param string $directoryPath The directory path
     *
     * @return $this
     */
    public function setDirectoryPath (string $directoryPath): self {
        $this->directoryPath = $directoryPath;
        return $this;
    }

    /**
     * True if search in subdirectories too, else False
     *
     * @return bool True if search in subdirectories too, else False
     */
    public function isProcessSubDirectories (): bool {
        return $this->processSubDirectories;
    }
    /**
     * Set if search in subdirectories too
     *
     * @param bool $processSubDirectories True if search in subdirectories too, else False
     *
     * @return $this
     */
    public function setProcessSubDirectories (bool $processSubDirectories): self {
        $this->processSubDirectories = $processSubDirectories;
        return $this;
    }

    /**
     * @inheritDoc
     */
    public function getCommands (): array {
        return static::getClassesOfDirectory($this->getDirectoryPath(), $this->isProcessSubDirectories());
    }
    /**
     * Get spot's classes of a directory
     *
     * @param string $directoryPath         The directory path
     * @param bool   $processSubDirectories True if search in subdirectories too, else False
     *
     * @return Command[] Spot's classes
     */
    private static function getClassesOfDirectory (string $directoryPath, bool $processSubDirectories): array {
        $classes = [];

        $directoryIterator = new FilesystemIterator($directoryPath);
        foreach ($directoryIterator as $fileInfo) {
            if ($fileInfo->isDot() || mb_substr($fileInfo->getFilename(), 0, 1) === '.') {
                continue;
            }
            if ($fileInfo->isDir()) {
                if ($processSubDirectories) {
                    /** @noinspection PhpConditionAlreadyCheckedInspection */
                    $classes = array_merge(
                        $classes,
                        static::getClassesOfDirectory($fileInfo->getPathname(), $processSubDirectories)
                    );
                }
                continue;
            }

            $spotClass = AutoDiscoverySpotClass::createFromFile($fileInfo->getPathname());
            $classPath = realpath($spotClass->getName());
            if (array_key_exists($classPath, $classes)) {
                continue;
            }

            try {
                $class = new ReflectionClass($spotClass->getName());
                if ($class->isAbstract()
                    || $class->isInterface()
                    || $class->isTrait()
                    || !is_subclass_of($class->getName(), Command::class)
                ) {
                    continue;
                }

                $classes[$classPath] = $class->newInstance();
            }
            catch (ReflectionException $exception) {
                continue;
            }
        }

        return $classes;
    }
}